# -*- coding: utf-8 -*-

import pika
import os
import json
import itertools


class ZZNBus(object):

    def __init__(self, url, rpcExchange='RPC', eventExchange='EVT'):
        self._rpcExchange = rpcExchange
        self._eventExchange = eventExchange
        self._conn = pika.BlockingConnection(pika.URLParameters(url))
        self._chann = self._conn.channel()
        self._queue = self._chann.queue_declare(exclusive=True).method.queue
        self._results = {}
        self._ctag = self._chann.basic_consume(
            self.consumer, queue=self._queue, no_ack=True
        )
        self._chann.add_on_return_callback(self.on_return)
        self._chann.add_on_cancel_callback(self.on_cancel)
        self._chann.add_on_close_callback(self.on_close)
        self._ctag = self._chann.basic_consume(
            self.consumer, queue=self._queue, no_ack=True
        )
        self.rpc = RpcNamespace(self)

    _idCounter = itertools.count()

    def nextCorrelationId(self):
        return 'req-%d-%s-%d' % (
            os.getpid(), id(self), self._idCounter.next()
        )

    def on_cancel(self, *args, **kwds):
        print 'CANCEL', args, kwds

    def on_return(self, *args, **kwds):
        print 'RETURN', args, kwds

    def on_close(self, *args, **kwds):
        print 'CLOSE', args, kwds

    def consumer(self, ch, method, props, body):
        #print 'CB', method, props, body
        #self.chann.stop_consuming()
        self._results[props.correlation_id] = json.loads(body)

    def call(self, requests, callback=None):
        results = {}
        reqmap = {}
        for request in requests:
            request['id'] = cid = self.nextCorrelationId()
            reqmap[cid] = request
            self._chann.basic_publish(
                exchange=self._rpcExchange,
                routing_key=request['method'],
                properties=pika.BasicProperties(
                    reply_to=self._queue,
                    correlation_id=cid,
                ),
                body=json.dumps(request),
                mandatory=True,
                immediate=False,
            )
        while True:
            for cid, res in self._results.items():
                if callback is not None:
                    callback(reqmap[cid], res)
                results[cid] = res
            self._results.clear()
            if len(results) == len(requests):
                break
            self._conn.process_data_events()
        return results

    def __call__(self, request):
        res = self.call([request])
        return res.items()[0][1]

    def close(self):
        self._chann.close()
        self._conn.close()



class RpcNamespace(object):

    def __init__(self, bus, name=None):
        self.__bus = bus
        self.__name = name

    def __getattr__(self, method):
        if self.__name:
            return RpcNamespace(self.__bus, '%s.%s' % (self.__name, method))
        else:
            return RpcNamespace(self.__bus, method)

    def __call__(self, *args, **kwds):
        if args:
            assert not kwds
            params = list(args)
        elif kwds:
            params = [kwds]
        else:
            params = []
        request = dict(
            method=self.__name,
            params=params
        )
        res = self.__bus(request)
        if res.get('error'):
            raise Exception(res['error'])
        else:
            return res.get('result')


def xmain():
    import sys
    import argparse
    from pprint import pprint
    PASS = os.environ['MQ_PASSWORD']
    rpc = RPC("amqp://%s:%s@mq.is.uh.in/dat" % (os.getlogin(), PASS))
    requests = [json.load(open(fn)) for fn in sys.argv[1:]]
    result = rpc.call(requests)
    print json.dumps(result, indent=2, sort_keys=True)


def main(histfile='~/.rpc-shell-history'):
    import code
    import os
    import sys
    import pprint
    import rlcompleter
    import readline
    import atexit
    sys.displayhook = pprint.pprint
    url = "amqp://%s:%s@mq.is.uh.in/dat" % (
        os.getlogin(), os.environ['MQ_PASSWORD'])
    print url
    bus = ZZNBus(url)
    namespace = dict(bus=bus)
    readline.set_completer(rlcompleter.Completer(namespace).complete)
    readline.parse_and_bind("tab: complete")
    if histfile:
        histfile = os.path.expanduser(histfile)
        if os.path.isfile(histfile):
            readline.read_history_file(histfile)
        def savehist():
            readline.set_history_length(100)
            readline.write_history_file(histfile)
        atexit.register(savehist)
    con = code.InteractiveConsole(namespace)
    con.resetbuffer()
    con.interact(banner='''
bus = ZZNBus instance (connected)
''')
